# encoding: utf-8
"""
Shader and ShaderProgram wrapper classes for vertex and fragment shaders used
in Interactive Data Visualization
"""

import contextlib
import ctypes
import os
from collections import OrderedDict

import traitlets
import yaml
from OpenGL import GL

from yt.units.yt_array import YTQuantity
from yt.utilities.exceptions import (
    YTInvalidShaderType,
    YTUnknownUniformKind,
    YTUnknownUniformSize,
)

from .opengl_support import GLValue, num_to_const


class ShaderProgram:
    """
    Wrapper class that compiles and links vertex and fragment shaders
    into a shader program.

    Parameters
    ----------

    vertex_shader : string
        or :class:`yt.visualization.volume_rendering.shader_objects.VertexShader`
        The vertex shader used in the Interactive Data Visualization pipeline.

    fragment_shader : string
        or :class:`yt.visualization.volume_rendering.shader_objects.FragmentShader`
        The fragment shader used in the Interactive Data Visualization pipeline.
    """

    def __init__(self, vertex_shader=None, fragment_shader=None):
        # Don't allow just one.  Either neither or both.
        if vertex_shader is None and fragment_shader is None:
            pass
        elif None not in (vertex_shader, fragment_shader):
            self.link(vertex_shader, fragment_shader)
        else:
            raise RuntimeError
        self._uniform_funcs = OrderedDict()

    def link(self, vertex_shader, fragment_shader):
        # There are more types of shaders, but for now we only allow v&f.
        self.program = GL.glCreateProgram()
        if not isinstance(vertex_shader, Shader):
            vertex_shader = Shader(source=vertex_shader)
        if not isinstance(fragment_shader, Shader):
            fragment_shader = Shader(source=fragment_shader)
        self.vertex_shader = vertex_shader
        self.fragment_shader = fragment_shader
        GL.glAttachShader(self.program, vertex_shader.shader)
        GL.glAttachShader(self.program, fragment_shader.shader)
        GL.glLinkProgram(self.program)
        result = GL.glGetProgramiv(self.program, GL.GL_LINK_STATUS)
        if not result:
            raise RuntimeError(GL.glGetProgramInfoLog(self.program))
        vertex_shader.delete_shader()
        fragment_shader.delete_shader()
        self.introspect()

    def introspect(self):
        if self.program is None:
            raise RuntimeError
        # First get all of the uniforms
        self.uniforms = {}
        self.attributes = {}

        if not bool(GL.glGetProgramInterfaceiv):
            return
        n_uniforms = GL.glGetProgramInterfaceiv(
            self.program, GL.GL_UNIFORM, GL.GL_ACTIVE_RESOURCES
        )

        for i in range(n_uniforms):
            name, size, gl_type = GL.glGetActiveUniform(self.program, i)
            gl_type = num_to_const[gl_type]
            self.uniforms[name.decode("utf-8")] = (size, gl_type)

        n_attrib = GL.glGetProgramInterfaceiv(
            self.program, GL.GL_PROGRAM_INPUT, GL.GL_ACTIVE_RESOURCES
        )
        length = ctypes.pointer(ctypes.c_int())
        size = ctypes.pointer(ctypes.c_int())
        gl_type = ctypes.pointer(ctypes.c_int())
        name = ctypes.create_string_buffer(256)
        for i in range(n_attrib):
            GL.glGetActiveAttrib(self.program, i, 256, length, size, gl_type, name)
            gl_const = num_to_const[gl_type[0]]
            self.attributes[name[: length[0]].decode("utf-8")] = (size[0], gl_const)

    def delete_program(self):
        if self.program is not None:
            GL.glDeleteProgram(self.program)
            self.program = None

    def _guess_uniform_func(self, value):
        # We make a best-effort guess.
        # This does NOT work with arrays of uniforms.
        # First look at the dtype kind.  Fortunately, this falls into either
        # 'f' or 'i', which matches nicely with OpenGL.
        # Note that in some implementations, it seems there is also a 'd' type,
        # but we will not be using that here.
        if isinstance(value, int):
            return GL.glUniform1i
        elif isinstance(value, (YTQuantity, float)):
            return GL.glUniform1f
        else:
            kind = value.dtype.kind
        if kind not in "if":
            raise YTUnknownUniformKind(kind)
        if value.ndim == 0:
            return {"f": GL.glUniform1f, "i": GL.glUniform1i}[kind]
        elif value.ndim == 1:
            if value.size > 4:
                raise YTUnknownUniformSize(value.size)
            func = self._set_scalar_uniform(kind, value.size)
        elif value.ndim == 2:
            if value.shape[0] != value.shape[1]:
                raise YTUnknownUniformSize(value.shape)
            func = self._set_matrix_uniform(kind, value.shape)
        else:
            raise YTUnknownUniformSize(value.shape)
        return func

    def _set_scalar_uniform(self, kind, size_spec):
        gl_func = getattr(GL, f"glUniform{size_spec}{kind}v")

        def _func(location, value):
            return gl_func(location, 1, value)

        return _func

    def _set_matrix_uniform(self, kind, size_spec):
        assert size_spec[0] == size_spec[1]
        gl_func = getattr(GL, f"glUniformMatrix{size_spec[0]}{kind}v")

        def _func(location, value):
            return gl_func(location, 1, GL.GL_TRUE, value)

        return _func

    def _set_uniform(self, name, value):
        # We need to figure out how to pass it in.
        if name not in self._uniform_funcs:
            self._uniform_funcs[name] = self._guess_uniform_func(value)
        loc = GL.glGetUniformLocation(self.program, name)
        if loc < 0:
            return -1
        return self._uniform_funcs[name](loc, value)

    @contextlib.contextmanager
    def enable(self):
        GL.glUseProgram(self.program)
        self.vertex_shader.setup_blend()
        self.fragment_shader.setup_blend()
        yield self
        GL.glUseProgram(0)


class Shader(traitlets.HasTraits):
    """
    Creates a shader from source

    Parameters
    ----------

    source : str
        This can either be a string containing a full source of a shader,
        an absolute path to a source file or a filename of a shader
        residing in the ./shaders/ directory.

    """

    _shader = None
    source = traitlets.Any()
    shader_name = traitlets.CUnicode()
    info = traitlets.CUnicode()
    shader_type = traitlets.CaselessStrEnum(("vertex", "fragment"))
    blend_func = traitlets.Tuple(
        GLValue(), GLValue(), default_value=("src alpha", "dst alpha")
    )
    blend_equation = GLValue("func add")
    depth_test = GLValue("always")

    use_separate_blend = traitlets.Bool(False)
    blend_equation_separate = traitlets.Tuple(
        GLValue(), GLValue(), default_value=("none", "none")
    )
    blend_func_separate = traitlets.Tuple(
        GLValue(),
        GLValue(),
        GLValue(),
        GLValue(),
        default_value=("none", "none", "none", "none"),
    )

    def _get_source(self, source):
        if ";" in source:
            # This is probably safe, right?  Enh, probably.
            return source
        # What this does is concatenate multiple (if available) source files.
        # This gets around GLSL's composition issues, which means we can have
        # functions that get called at each step in a ray tracing process, for
        # instance, that can still share ray tracing code between multiple
        # files.
        if not isinstance(source, (tuple, list)):
            source = (source,)
        full_source = []
        for fn in source:
            if os.path.isfile(fn):
                sh_directory = ""
            else:
                sh_directory = os.path.join(os.path.dirname(__file__), "shaders")
            fn = os.path.join(sh_directory, fn)
            if not os.path.isfile(fn):
                raise YTInvalidShaderType(fn)
            full_source.append(open(fn, "r").read())
        return "\n\n".join(full_source)

    def compile(self, source=None, parameters=None):
        if source is None:
            source = self.source
            if source is None:
                raise RuntimeError
        if parameters is not None:
            raise NotImplementedError
        source = self._get_source(source)
        shader_type_enum = getattr(GL, f"GL_{self.shader_type.upper()}_SHADER")
        shader = GL.glCreateShader(shader_type_enum)
        # We could do templating here if we wanted.
        self.shader_source = source
        GL.glShaderSource(shader, source)
        GL.glCompileShader(shader)
        result = GL.glGetShaderiv(shader, GL.GL_COMPILE_STATUS)
        if not (result):
            raise RuntimeError(GL.glGetShaderInfoLog(shader))
        self._shader = shader

    def setup_blend(self):
        GL.glEnable(GL.GL_BLEND)
        if self.use_separate_blend:
            GL.glBlendEquationSeparate(*self.blend_equation_separate)
            GL.glBlendFuncSeparate(*self.blend_func_separate)
        else:
            GL.glBlendEquation(self.blend_equation)
            GL.glBlendFunc(*self.blend_func)
        GL.glEnable(GL.GL_DEPTH_TEST)
        GL.glDepthFunc(self.depth_test)

    @property
    def shader(self):
        if self._shader is None:
            self.compile()
        return self._shader

    def delete_shader(self):
        if None not in (self._shader, GL.glDeleteShader):
            GL.glDeleteShader(self._shader)
            self._shader = None

    def __del__(self):
        # This is not guaranteed to be called
        self.delete_shader()


class ShaderTrait(traitlets.TraitType):
    default_value = None
    info_text = "A shader (vertex or fragment)"

    def validate(self, obj, value):
        if isinstance(value, str):
            try:
                shader_type = self.metadata.get("shader_type", "vertex")
                shader_info = known_shaders[shader_type][value]
                shader_info.setdefault("shader_type", shader_type)
                shader_info["use_separate_blend"] = bool(
                    "blend_func_separate" in shader_info
                )
                shader_info.setdefault("shader_name", value)
                shader = Shader(**shader_info)
                return shader
            except KeyError:
                self.error(obj, value)
        elif isinstance(value, Shader):
            return value
        self.error(obj, value)


known_shaders = {}
component_shaders = {}
default_shader_combos = {}

# We'll load our shaders here from shaderlist.yaml
_shlist_fn = os.path.join(os.path.dirname(__file__), "shaders", "shaderlist.yaml")
if os.path.exists(_shlist_fn):
    with open(_shlist_fn, "r") as f:
        shader_info = yaml.load(f, yaml.SafeLoader)
    known_shaders.update(shader_info["shader_definitions"])
    component_shaders.update(shader_info["component_shaders"])
    default_shader_combos.update(
        {_: component_shaders[_].pop("default_value") for _ in component_shaders}
    )



import string
from collections import namedtuple

import numpy as np
import traitlets
from OpenGL import GL
from matplotlib.ft2font import LOAD_FORCE_AUTOHINT

from yt_idv.opengl_support import VertexArray, Texture2D, VertexAttribute
from yt_idv.scene_data.base_data import SceneData
from yt_idv.traitlets_support import FontTrait

Character = namedtuple(
    "Character", ["texture", "vbo_offset", "hori_advance", "vert_advance"]
)


class TextCharacters(SceneData):
    characters = traitlets.Dict(trait=traitlets.Instance(Character))
    name = "text_overlay"
    font = FontTrait("DejaVu Sans")
    font_size = traitlets.CInt(32)

    @traitlets.default("vertex_array")
    def _default_vertex_array(self):
        return VertexArray(name="char_info", each=6)

    def build_textures(self):
        # This doesn't check if the textures have already been built
        self.font.set_size(self.font_size, 200)
        chars = [ord(_) for _ in string.printable]
        tex_ids = GL.glGenTextures(len(chars))
        vert = []
        for i, (tex_id, char_code) in enumerate(zip(tex_ids, chars)):
            self.font.clear()
            self.font.set_text(chr(char_code), flags=LOAD_FORCE_AUTOHINT)
            self.font.draw_glyphs_to_bitmap(antialiased=True)
            glyph = self.font.load_char(char_code)
            x0, y0, x1, y1 = glyph.bbox
            bitmap = self.font.get_image().astype(">f4") / 255.0
            dx = 1.0 / bitmap.shape[0]
            dy = 1.0 / bitmap.shape[1]
            triangles = np.array(
                [
                    [x0, y1, 0.0 + dx / 2.0, 0.0 + dy / 2.0],
                    [x0, y0, 0.0 + dx / 2.0, 1.0 - dy / 2.0],
                    [x1, y0, 1.0 - dx / 2.0, 1.0 - dy / 2.0],
                    [x0, y1, 0.0 + dx / 2.0, 0.0 + dy / 2.0],
                    [x1, y0, 1.0 - dx / 2.0, 1.0 - dy / 2.0],
                    [x1, y1, 1.0 - dx / 2.0, 0.0 + dy / 2.0],
                ],
                dtype="<f4",
            )
            vert.append(triangles)
            texture = Texture2D(
                texture_name=tex_id, data=bitmap, boundary_x="clamp", boundary_y="clamp"
            )
            # I can't find information as to why horiAdvance is a
            # factor of 8 larger than the other factors.  I assume it
            # is referenced somewhere, but I cannot find it.
            self.characters[chr(char_code)] = Character(
                texture, i, glyph.horiAdvance / 8.0, glyph.vertAdvance
            )
        vert = np.concatenate(vert)
        self.vertex_array.attributes.append(
            VertexAttribute(name="quad_vertex", data=vert.astype("<f4"))
        )



import numpy as np
import traitlets
from OpenGL import GL

from yt_idv.scene_components.base_component import SceneComponent
from yt_idv.scene_data.mesh import MeshData


class MeshRendering(SceneComponent):
    name = "mesh_rendering"
    data = traitlets.Instance(MeshData)

    def draw(self, scene, program):
        GL.glDrawElements(GL.GL_TRIANGLES, self.data.size, GL.GL_UNSIGNED_INT, None)

    def _set_uniforms(self, scene, shader_program):
        projection_matrix = scene.camera.projection_matrix
        view_matrix = scene.camera.view_matrix
        model_to_clip = np.dot(projection_matrix, view_matrix)
        shader_program._set_uniform("model_to_clip", model_to_clip)



import contextlib

import numpy as np
import traitlets
import traittypes

from yt_idv.traitlets_support import YTPositionTrait, ndarray_ro, ndarray_shape


class BaseCamera(traitlets.HasTraits):
    """Camera object used in the Interactive Data Visualization

    Parameters
    ----------

    position : :obj:`!iterable`, or 3 element array in code_length
        The initial position of the camera.
    focus : :obj:`!iterable`, or 3 element array in code_length
        A point in space that the camera is looking at.
    up : :obj:`!iterable`, or 3 element array in code_length
        The 'up' direction for the camera.
    fov : float, optional
        An angle defining field of view in degrees.
    near_plane : float, optional
        The distance to the near plane of the perspective camera.
    far_plane : float, optional
        The distance to the far plane of the perspective camera.
    aspect_ratio: float, optional
        The ratio between the height and the width of the camera's fov.

    """

    # We have to be careful about some of these, as it's possible in-place
    # operations won't trigger our observation.
    position = YTPositionTrait([0.0, 0.0, 1.0])
    focus = YTPositionTrait([0.0, 0.0, 0.0])
    up = traittypes.Array(np.array([0.0, 0.0, 1.0])).valid(
        ndarray_shape(3), ndarray_ro()
    )
    fov = traitlets.Float(45.0)
    near_plane = traitlets.Float(0.001)
    far_plane = traitlets.Float(20.0)
    aspect_ratio = traitlets.Float(8.0 / 6.0)

    projection_matrix = traittypes.Array(np.zeros((4, 4))).valid(
        ndarray_shape(4, 4), ndarray_ro()
    )
    view_matrix = traittypes.Array(np.zeros((4, 4))).valid(
        ndarray_shape(4, 4), ndarray_ro()
    )
    orientation = traittypes.Array(np.zeros(4)).valid(ndarray_shape(4), ndarray_ro())

    held = traitlets.Bool(False)

    @contextlib.contextmanager
    def hold_traits(self, func):
        # for some reason, hold_trait_notifications doesn't seem to work here.
        # So, we use this to block.  We also do not want to pass the
        # notifications once completed.
        if not self.held:
            self.held = True
            func()
            self.held = False
        yield

    @traitlets.default("up")
    def _default_up(self):
        return np.array([0.0, 1.0, 0.0])

    @traitlets.observe(
        "position",
        "focus",
        "up",
        "fov",
        "near_plane",
        "far_plane",
        "aspect_ratio",
        "orientation",
    )
    def compute_matrices(self, change=None):
        """Regenerate all position, view and projection matrices of the camera."""
        with self.hold_traits(self._compute_matrices):
            pass

    def update_orientation(self, start_x, start_y, end_x, end_y):
        """Change camera orientation matrix using delta of mouse's cursor position

        Parameters
        ----------

        start_x : float
            initial cursor position in x direction
        start_y : float
            initial cursor position in y direction
        end_x : float
            final cursor position in x direction
        end_y : float
            final cursor position in y direction

        """
        pass



import traitlets

from yt_idv.constants import FULLSCREEN_QUAD
from yt_idv.opengl_support import Texture1D, VertexArray, VertexAttribute
from yt_idv.scene_data.base_data import SceneData


class RGBAData(SceneData):
    name = "rgba_data"
    colormap_texture = traitlets.Instance(Texture1D)

    @traitlets.default("vertex_array")
    def _default_vertex_array(self):
        va = VertexArray(name="tri", each=6)
        fq = FULLSCREEN_QUAD.reshape((6, 3), order="C")
        va.attributes.append(VertexAttribute(name="vertexPosition_modelspace", data=fq))
        return va

    def add_data(self, lines):
        assert lines.shape[1] == 4
        self.colormap_texture = Texture1D(boundary_x="clamp", data=lines)



import matplotlib.font_manager
import numpy as np
import traitlets


class YTPositionTrait(traitlets.TraitType):
    default_value = None
    info_text = "A position in code_length"

    def validate(self, obj, value):
        if isinstance(value, (tuple, list)):
            value = np.array(value)
        if hasattr(value, "in_units"):
            value = value.in_units("unitary").d
        if not isinstance(value, np.ndarray):
            self.error(obj, value)
        return value.astype("f4")


def ndarray_shape(*dimensions):
    # http://traittypes.readthedocs.io/en/latest/api_documentation.html
    def validator(trait, value):
        if value.shape != dimensions:
            raise traitlets.TraitError(
                "Expected an of shape %s and got and array with shape %s"
                % (dimensions, value.shape)
            )
        else:
            return value

    return validator


def ndarray_ro():
    def validator(trait, value):
        if value.flags["WRITEABLE"]:
            value = value.copy()
            value.flags["WRITEABLE"] = False
        return value

    return validator


class FontTrait(traitlets.TraitType):
    info_text = "A font instance from matplotlib"

    def validate(self, obj, value):
        if isinstance(value, str):
            try:
                font_fn = matplotlib.font_manager.findfont(value)
                value = matplotlib.font_manager.get_font(font_fn)
            except FileNotFoundError:
                self.error(obj, value)
        return value



from math import floor, ceil

import numpy as np
import traitlets
from OpenGL import GL

from yt_idv.opengl_support import TransferFunctionTexture
from yt_idv.scene_components.base_component import SceneComponent
from yt_idv.scene_data.block_collection import BlockCollection
from yt_idv.shader_objects import component_shaders


class BlockRendering(SceneComponent):
    """
    A class that renders block data.  It may do this in one of several ways,
    including mesh outline.  This allows us to render a single collection of
    blocks multiple times in a single scene and to separate out the memory
    handling from the display.
    """

    name = "block_rendering"
    data = traitlets.Instance(BlockCollection)
    box_width = traitlets.CFloat(0.1)
    transfer_function = traitlets.Instance(TransferFunctionTexture)
    tf_min = traitlets.CFloat(0.0)
    tf_max = traitlets.CFloat(1.0)
    tf_log = traitlets.Bool(True)

    priority = 10

    def render_gui(self, imgui, renderer):
        changed = super(BlockRendering, self).render_gui(imgui, renderer)
        _, self.cmap_log = imgui.checkbox("Take log", self.cmap_log)
        changed = changed or _
        _, cmap_index = imgui.listbox(
            "Colormap", _cmaps.index(self.colormap.colormap_name), _cmaps
        )
        if _:
            self.colormap.colormap_name = _cmaps[cmap_index]
        changed = changed or _
        # Now, shaders
        shader_combos = list(sorted(component_shaders[self.name]))
        descriptions = [
            component_shaders[self.name][_]["description"] for _ in shader_combos
        ]
        selected = shader_combos.index(self.render_method)
        _, shader_ind = imgui.listbox("Shader", selected, descriptions)
        if _:
            self.render_method = shader_combos[shader_ind]
        changed = changed or _

        # Now for the transfer function stuff
        imgui.image_button(
            self.transfer_function.texture_name, 256, 32, frame_padding=0
        )
        imgui.text("Right click and drag to change")
        update = False
        data = self.transfer_function.data.astype("f4") / 255
        for i, c in enumerate("rgba"):
            imgui.plot_lines(
                f"## {c}",
                data[:, 0, i].copy(),
                scale_min=0.0,
                scale_max=1.0,
                graph_size=(256, 32),
            )
            if imgui.is_item_hovered() and imgui.is_mouse_dragging(2):
                update = True
                dx, dy = renderer.io.mouse_delta
                dy = -dy
                mi = imgui.get_item_rect_min()
                ma = imgui.get_item_rect_max()
                x, y = renderer.io.mouse_pos
                x = x - mi.x
                y = (ma.y - mi.y) - (y - mi.y)
                xb1 = floor(min(x + dx, x) * data.shape[0] / (ma.x - mi.x))
                xb2 = ceil(max(x + dx, x) * data.shape[0] / (ma.x - mi.x))
                yv1 = y / (ma.y - mi.y)
                yv2 = (y + dy) / (ma.y - mi.y)
                yv1, yv2 = (max(min(_, 1.0), 0.0) for _ in (yv1, yv2))
                if dx < 0:
                    yv2, yv1 = yv1, yv2
                    xb1 -= 1
                elif dx > 0:
                    xb2 += 1
                xb1 = max(0, xb1)
                xb2 = min(255, xb2)
                if renderer.io.key_shift:
                    yv1 = yv2 = 1.0
                elif renderer.io.key_ctrl:
                    yv1 = yv2 = 0.0
                data[xb1:xb2, 0, i] = np.mgrid[yv1 : yv2 : (xb2 - xb1) * 1j]
        if update:
            self.transfer_function.data = (data * 255).astype("u1")

    @traitlets.default("transfer_function")
    def _default_transfer_function(self):
        tf = TransferFunctionTexture(data=np.ones((256, 1, 4), dtype="u1") * 255)
        return tf

    def draw(self, scene, program):
        each = self.data.vertex_array.each
        with self.transfer_function.bind(target=2):
            for tex_ind, tex, bitmap_tex in self.data.viewpoint_iter(scene.camera):
                with tex.bind(target=0):
                    with bitmap_tex.bind(target=1):
                        GL.glDrawArrays(GL.GL_TRIANGLES, tex_ind * each, each)

    def _set_uniforms(self, scene, shader_program):
        cam = scene.camera
        shader_program._set_uniform("projection", cam.projection_matrix)
        shader_program._set_uniform("modelview", cam.view_matrix)
        shader_program._set_uniform(
            "viewport", np.array(GL.glGetIntegerv(GL.GL_VIEWPORT), dtype="f4")
        )
        shader_program._set_uniform("camera_pos", cam.position)
        shader_program._set_uniform("box_width", self.box_width)
        shader_program._set_uniform("ds_tex", 0)
        shader_program._set_uniform("bitmap_tex", 1)
        shader_program._set_uniform("tf_tex", 2)
        shader_program._set_uniform("tf_min", self.tf_min)
        shader_program._set_uniform("tf_max", self.tf_max)
        shader_program._set_uniform("tf_log", float(self.tf_log))


_cmaps = ["arbre", "viridis", "magma", "doom"]



import numpy as np
import traitlets
from OpenGL import GL

from yt_idv.scene_annotations.base_annotation import SceneAnnotation
from yt_idv.scene_data.box import BoxData


class BoxAnnotation(SceneAnnotation):
    name = "box_outline"
    data = traitlets.Instance(BoxData)
    box_width = traitlets.CFloat(0.05)
    box_alpha = traitlets.CFloat(1.0)

    def draw(self, scene, program):
        each = self.data.vertex_array.each
        GL.glDrawArrays(GL.GL_TRIANGLES, 0, each)

    def render_gui(self, imgui, renderer):
        changed = super(BoxAnnotation, self).render_gui(imgui, renderer)
        _, bw = imgui.slider_float("Width", self.box_width, 0.001, 0.250)
        if _:
            self.box_width = bw
        changed = changed or _
        _, ba = imgui.slider_float("Alpha", self.box_alpha, 0.0, 1.0)
        if _:
            self.box_alpha = ba
        changed = changed or _
        return changed

    def _set_uniforms(self, scene, shader_program):
        cam = scene.camera
        shader_program._set_uniform("projection", cam.projection_matrix)
        shader_program._set_uniform("modelview", cam.view_matrix)
        shader_program._set_uniform(
            "viewport", np.array(GL.glGetIntegerv(GL.GL_VIEWPORT), dtype="f4")
        )
        shader_program._set_uniform("camera_pos", cam.position)
        shader_program._set_uniform("box_width", self.box_width)
        shader_program._set_uniform("box_alpha", self.box_alpha)



# encoding: utf-8
"""
Shader and ShaderProgram wrapper classes for vertex and fragment shaders used
in Interactive Data Visualization
"""

# ----------------------------------------------------------------------------
# Copyright (c) 2016, yt Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
# ----------------------------------------------------------------------------

# This is a part of the experimental Interactive Data Visualization

from contextlib import contextmanager

import matplotlib.cm as cm
import numpy as np
import traitlets
import traittypes
from OpenGL import GL

from contextlib import ExitStack

# Set up a mapping from numbers to names
from yt.utilities.math_utils import get_translate_matrix, get_scale_matrix

from yt_idv.constants import bbox_vertices

const_types = (
    GL.constant.IntConstant,
    GL.constant.LongConstant,
    GL.constant.FloatConstant,
)

num_to_const = {}
for i in dir(GL):
    if i.startswith("GL_"):
        v = getattr(GL, i)
        if not isinstance(v, const_types):
            continue
        num_to_const[v.real] = v

_coersion_funcs = {
    "FLOAT": float,
    "DOUBLE": float,
    "INT": int,
    "UNSIGNED": int,
    "BOOL": bool,
}

_shapes = {
    "MAT2": (2, 2),
    "MAT3": (3, 3),
    "MAT4": (4, 4),
    "MAT2x3": (2, 3),
    "MAT2x4": (2, 4),
    "MAT3x2": (3, 2),
    "MAT3x4": (3, 4),
    "MAT4x2": (4, 2),
    "MAT4x3": (4, 3),
    "VEC2": (2,),
    "VEC3": (3,),
    "VEC4": (4,),
}

# PyOpenGL has the reverse mapping for this
gl_to_np = {
    "FLOAT": "f",
    "DOUBLE": "d",
    "INT": "i",
    "UNSIGNED": "I",
    "BOOL": "b",
}

np_to_gl = {
    "float32": GL.GL_FLOAT,
    "uint32": GL.GL_UNSIGNED_INT,
    "uint8": GL.GL_UNSIGNED_BYTE,
}

TEX_CHANNELS = {
    "float32": {
        1: (GL.GL_FLOAT, GL.GL_R32F, GL.GL_RED),
        2: (GL.GL_FLOAT, GL.GL_RG32F, GL.GL_RG),
        3: (GL.GL_FLOAT, GL.GL_RGB32F, GL.GL_RGB),
        4: (GL.GL_FLOAT, GL.GL_RGBA32F, GL.GL_RGBA),
    },
    "uint8": {
        1: (GL.GL_UNSIGNED_BYTE, GL.GL_R8, GL.GL_RED),
        2: (GL.GL_UNSIGNED_BYTE, GL.GL_RG8, GL.GL_RG),
        3: (GL.GL_UNSIGNED_BYTE, GL.GL_RGB8, GL.GL_RGB),
        4: (GL.GL_UNSIGNED_BYTE, GL.GL_RGBA8, GL.GL_RGBA),
    },
    "uint32": {
        1: (GL.GL_UNSIGNED_INT, GL.GL_R32UI, GL.GL_RED),
        2: (GL.GL_UNSIGNED_INT, GL.GL_RG32UI, GL.GL_RG),
        3: (GL.GL_UNSIGNED_INT, GL.GL_RGB32UI, GL.GL_RGB),
        4: (GL.GL_UNSIGNED_INT, GL.GL_RGBA32UI, GL.GL_RGBA),
    },
}


def coerce_uniform_type(val, gl_type):
    # gl_type here must be in const_types
    if not isinstance(gl_type, const_types):
        gl_type = num_to_const[gl_type]
    # Now we can get down to business!
    spec = gl_type.name.split("_")[1:]  # Strip out the GL_
    # We know what to do with:
    #   FLOAT DOUBLE INT UNSIGNED BOOL
    # We can ignore:
    #    SAMPLER IMAGE
    if "SAMPLER" in spec or "IMAGE" in spec:
        # Do nothing to these, and let PyOpenGL handle it
        return val
    if len(spec) == 1 or spec == ["UNSIGNED", "INT"]:
        return _coersion_funcs[spec[0]](val)
    # We need to figure out if it's a matrix, a vector, etc.
    shape = _shapes[spec[-1]]
    dtype = gl_to_np[spec[0]]
    val = np.asanyarray(val, dtype=dtype)
    val.shape = shape
    return val


class TextureBoundary(traitlets.TraitType):
    default_value = GL.GL_CLAMP_TO_EDGE
    info_text = "A boundary type of mirror, clamp, or repeat"

    def validate(self, obj, value):
        if isinstance(value, str):
            try:
                return {
                    "clamp": GL.GL_CLAMP_TO_EDGE,
                    "mirror": GL.GL_MIRRORED_REPEAT,
                    "repeat": GL.GL_REPEAT,
                }[value.lower()]
            except KeyError:
                self.error(obj, value)
        elif value in (GL.GL_CLAMP_TO_EDGE, GL.GL_MIRRORED_REPEAT, GL.GL_REPEAT):
            return value
        self.error(obj, value)


class GLValue(traitlets.TraitType):
    default_value = GL.GL_NONE
    info_text = "An OpenGL constant."

    def validate(self, obj, value):
        # This will convert lower to upper and spaces to _ and also preprend
        # GL_ if needed.
        if isinstance(value, str):
            if not value.startswith("GL"):
                value = f"GL_{value}"
            value = getattr(GL, value.upper().replace(" ", "_"), None)
            if value is None:
                self.error(obj, value)
        return value


TEX_TARGETS = {i: getattr(GL, f"GL_TEXTURE{i}") for i in range(10)}


class Texture(traitlets.HasTraits):
    texture_name = traitlets.CInt(-1)
    data = traittypes.Array(None, allow_none=True)
    channels = GLValue("r32f")
    min_filter = GLValue("linear")
    mag_filter = GLValue("linear")

    @traitlets.default("texture_name")
    def _default_texture_name(self):
        return GL.glGenTextures(1)

    @contextmanager
    def bind(self, target=0):
        _ = GL.glActiveTexture(TEX_TARGETS[target])
        _ = GL.glBindTexture(self.dim_enum, self.texture_name)
        yield
        _ = GL.glActiveTexture(TEX_TARGETS[target])
        GL.glBindTexture(self.dim_enum, 0)


class Texture1D(Texture):
    boundary_x = TextureBoundary()
    dims = 1
    dim_enum = GLValue("texture 1d")

    @traitlets.observe("data")
    def _set_data(self, change):
        with self.bind():
            data = change["new"]
            if len(data.shape) == 2:
                channels = data.shape[-1]
            else:
                channels = 1
            dx = data.shape[0]
            gl_type, type1, type2 = TEX_CHANNELS[data.dtype.name][channels]
            GL.glPixelStorei(GL.GL_UNPACK_ALIGNMENT, 1)
            if not isinstance(change["old"], np.ndarray):
                GL.glTexStorage1D(GL.GL_TEXTURE_1D, 1, type1, dx)
            GL.glTexSubImage1D(GL.GL_TEXTURE_1D, 0, 0, dx, type2, gl_type, data)
            GL.glTexParameterf(GL.GL_TEXTURE_1D, GL.GL_TEXTURE_WRAP_S, self.boundary_x)
            GL.glTexParameteri(
                GL.GL_TEXTURE_1D, GL.GL_TEXTURE_MIN_FILTER, self.min_filter
            )
            GL.glTexParameteri(
                GL.GL_TEXTURE_1D, GL.GL_TEXTURE_MAG_FILTER, self.mag_filter
            )
            GL.glGenerateMipmap(GL.GL_TEXTURE_1D)


class ColormapTexture(Texture1D):
    colormap_name = traitlets.CUnicode()

    def __init__(self, *args, **kwargs):
        # Override...
        kwargs["boundary_x"] = "clamp"
        super(ColormapTexture, self).__init__(*args, **kwargs)

    @traitlets.validate("colormap_name")
    def _validate_name(self, proposal):
        if proposal["value"] not in cm.cmap_d:
            raise traitlets.TraitError(
                "Colormap name needs to be known by" "matplotlib"
            )
        return proposal["value"]

    @traitlets.observe("colormap_name")
    def _observe_colormap_name(self, change):
        cmap = cm.get_cmap(change["new"])
        cmap_vals = np.array(cmap(np.linspace(0, 1, 256)), dtype="f4")
        self.data = cmap_vals


class Texture2D(Texture):
    boundary_x = TextureBoundary()
    boundary_y = TextureBoundary()
    dims = 2
    channels = 1
    dim_enum = GLValue("texture 2d")

    @traitlets.observe("data")
    def _set_data(self, change):
        with self.bind():
            data = change["new"]
            if len(data.shape) == 3:
                channels = data.shape[-1]
            else:
                channels = 1
            self.channels = channels
            dx, dy = data.shape[:2]
            gl_type, type1, type2 = TEX_CHANNELS[data.dtype.name][channels]
            GL.glPixelStorei(GL.GL_UNPACK_ALIGNMENT, 1)
            if not isinstance(change["old"], np.ndarray):
                GL.glTexStorage2D(GL.GL_TEXTURE_2D, 1, type1, dx, dy)
            GL.glTexSubImage2D(
                GL.GL_TEXTURE_2D, 0, 0, 0, dx, dy, type2, gl_type, data.swapaxes(0, 1)
            )
            GL.glTexParameterf(GL.GL_TEXTURE_2D, GL.GL_TEXTURE_WRAP_S, self.boundary_x)
            GL.glTexParameterf(GL.GL_TEXTURE_2D, GL.GL_TEXTURE_WRAP_T, self.boundary_y)
            GL.glTexParameteri(
                GL.GL_TEXTURE_2D, GL.GL_TEXTURE_MIN_FILTER, self.min_filter
            )
            GL.glTexParameteri(
                GL.GL_TEXTURE_2D, GL.GL_TEXTURE_MAG_FILTER, self.mag_filter
            )
            GL.glGenerateMipmap(GL.GL_TEXTURE_2D)


class TransferFunctionTexture(Texture2D):
    def __init__(self, *args, **kwargs):
        kwargs["boundary_x"] = "clamp"
        kwargs["boundary_y"] = "clamp"
        super(TransferFunctionTexture, self).__init__(*args, **kwargs)


class DepthBuffer(Texture2D):
    def create_texture(self, w, h):
        with self.bind():
            GL.glTexImage2D(
                GL.GL_TEXTURE_2D,
                0,
                GL.GL_DEPTH_COMPONENT24,
                w,
                h,
                0,
                GL.GL_DEPTH_COMPONENT,
                GL.GL_FLOAT,
                None,
            )
            GL.glTexParameterf(GL.GL_TEXTURE_2D, GL.GL_TEXTURE_WRAP_S, self.boundary_x)
            GL.glTexParameterf(GL.GL_TEXTURE_2D, GL.GL_TEXTURE_WRAP_T, self.boundary_y)


class Texture3D(Texture):
    boundary_x = TextureBoundary()
    boundary_y = TextureBoundary()
    boundary_z = TextureBoundary()
    dims = 3
    dim_enum = GLValue("texture 3d")

    @traitlets.observe("data")
    def _set_data(self, change):
        with self.bind():
            data = change["new"]
            if len(data.shape) == 4:
                channels = data.shape[-1]
            else:
                channels = 1
            dx, dy, dz = data.shape[:3]
            gl_type, type1, type2 = TEX_CHANNELS[data.dtype.name][channels]
            GL.glPixelStorei(GL.GL_UNPACK_ALIGNMENT, 1)
            GL.glTexParameterf(GL.GL_TEXTURE_3D, GL.GL_TEXTURE_WRAP_S, self.boundary_x)
            GL.glTexParameterf(GL.GL_TEXTURE_3D, GL.GL_TEXTURE_WRAP_T, self.boundary_y)
            GL.glTexParameterf(GL.GL_TEXTURE_3D, GL.GL_TEXTURE_WRAP_R, self.boundary_z)
            if not isinstance(change["old"], np.ndarray):
                GL.glTexStorage3D(GL.GL_TEXTURE_3D, 1, type1, dx, dy, dz)
            GL.glTexSubImage3D(
                GL.GL_TEXTURE_3D, 0, 0, 0, 0, dx, dy, dz, type2, gl_type, data.T
            )
            GL.glTexParameteri(
                GL.GL_TEXTURE_3D, GL.GL_TEXTURE_MIN_FILTER, self.min_filter
            )
            GL.glTexParameteri(
                GL.GL_TEXTURE_3D, GL.GL_TEXTURE_MAG_FILTER, self.mag_filter
            )
            GL.glGenerateMipmap(GL.GL_TEXTURE_3D)


class VertexAttribute(traitlets.HasTraits):
    name = traitlets.CUnicode("attr")
    id = traitlets.CInt(-1)
    data = traittypes.Array(None, allow_none=True)
    each = traitlets.CInt(-1)
    opengl_type = traitlets.CInt(GL.GL_FLOAT)

    @traitlets.default("id")
    def _id_default(self):
        return GL.glGenBuffers(1)

    @contextmanager
    def bind(self, program=None):
        loc = -1
        if program is not None:
            loc = GL.glGetAttribLocation(program.program, self.name)
            if loc >= 0:
                _ = GL.glEnableVertexAttribArray(loc)
        _ = GL.glBindBuffer(GL.GL_ARRAY_BUFFER, self.id)
        if loc >= 0:
            GL.glVertexAttribPointer(loc, self.each, self.opengl_type, False, 0, None)
        yield
        if loc >= 0:
            GL.glDisableVertexAttribArray(loc)
        GL.glBindBuffer(GL.GL_ARRAY_BUFFER, 0)

    @traitlets.observe("data")
    def _set_data(self, change):
        arr = change["new"]
        self.each = arr.shape[-1]
        self.opengl_type = np_to_gl[arr.dtype.name]
        with self.bind():
            GL.glBufferData(GL.GL_ARRAY_BUFFER, arr.nbytes, arr, GL.GL_STATIC_DRAW)


class VertexArray(traitlets.HasTraits):
    name = traitlets.CUnicode("vertex")
    id = traitlets.CInt(-1)
    indices = traittypes.Array(None, allow_none=True)
    index_id = traitlets.CInt(-1)
    attributes = traitlets.List(trait=traitlets.Instance(VertexAttribute))
    each = traitlets.CInt(-1)

    @traitlets.default("id")
    def _id_default(self):
        return GL.glGenVertexArrays(1)

    @contextmanager
    def bind(self, program=None):
        GL.glBindVertexArray(self.id)
        if self.index_id != -1:
            GL.glBindBuffer(GL.GL_ELEMENT_ARRAY_BUFFER, self.index_id)
        # We only bind the attributes if we have a program too
        if program is None:
            attrs = []
        else:
            attrs = self.attributes
        with ExitStack() as stack:
            _ = [stack.enter_context(_.bind(program)) for _ in attrs]
            yield
        if self.index_id != -1:
            GL.glBindBuffer(GL.GL_ELEMENT_ARRAY_BUFFER, 0)
        GL.glBindVertexArray(0)

    @traitlets.observe("indices")
    def _set_indices(self, change):
        arr = change["new"]
        self.index_id = GL.glGenBuffers(1)
        GL.glBindBuffer(GL.GL_ELEMENT_ARRAY_BUFFER, self.index_id)
        GL.glBufferData(GL.GL_ELEMENT_ARRAY_BUFFER, arr.nbytes, arr, GL.GL_STATIC_DRAW)


class Framebuffer(traitlets.HasTraits):
    fb_id = traitlets.CInt(-1)
    rb_id = traitlets.CInt(-1)
    fb_tex = traitlets.Instance(Texture2D)
    db_tex = traitlets.Instance(DepthBuffer)
    viewport = traitlets.Tuple(
        traitlets.CInt(), traitlets.CInt(), traitlets.CInt(), traitlets.CInt()
    )
    initialized = traitlets.Bool(False)

    @property
    def data(self):
        origin_x, origin_y, width, height = self.viewport
        with self.bind(clear=False):
            arr = GL.glReadPixels(0, 0, width, height, GL.GL_RGBA, GL.GL_FLOAT)
        return arr

    @property
    def depth_data(self):
        origin_x, origin_y, width, height = self.viewport
        with self.bind(clear=False):
            arr = GL.glReadPixels(
                0, 0, width, height, GL.GL_DEPTH_COMPONENT, GL.GL_FLOAT
            )
        return arr

    @traitlets.default("viewport")
    def _viewport_default(self):
        # origin_x, origin_y, width, height
        return tuple(GL.glGetIntegerv(GL.GL_VIEWPORT))

    @traitlets.observe("viewport")
    def _viewport_changed(self, change):
        # we just need to disable the initialized value here
        self.initalized = False

    @traitlets.default("fb_id")
    def _fb_id_default(self):
        return GL.glGenFramebuffers(1)

    @traitlets.default("rb_id")
    def _rb_id_default(self):
        return GL.glGenRenderbuffers(1)

    @traitlets.default("fb_tex")
    def _fb_tex_default(self):
        data = np.zeros((self.viewport[2], self.viewport[3], 4), "f4")
        return Texture2D(data=data, boundary_x="repeat", boundary_y="repeat")

    @traitlets.default("db_tex")
    def _db_tex_default(self):
        db = DepthBuffer(boundary_x="repeat", boundary_y="repeat")
        db.create_texture(self.viewport[2], self.viewport[3])
        return db

    @contextmanager
    def bind(self, clear=True):
        self.viewport = tuple(GL.glGetIntegerv(GL.GL_VIEWPORT))
        if not self.initialized:
            GL.glBindFramebuffer(GL.GL_FRAMEBUFFER, self.fb_id)
            GL.glBindRenderbuffer(GL.GL_RENDERBUFFER, self.rb_id)
            GL.glRenderbufferStorage(
                GL.GL_RENDERBUFFER,
                GL.GL_DEPTH_COMPONENT32F,
                self.viewport[2],
                self.viewport[3],
            )
            GL.glFramebufferRenderbuffer(
                GL.GL_FRAMEBUFFER,
                GL.GL_DEPTH_ATTACHMENT,
                GL.GL_RENDERBUFFER,
                self.rb_id,
            )

            GL.glFramebufferTexture2D(
                GL.GL_FRAMEBUFFER,
                GL.GL_COLOR_ATTACHMENT0,
                GL.GL_TEXTURE_2D,
                self.fb_tex.texture_name,
                0,  # mipmap level, normally 0
            )
            GL.glFramebufferTexture2D(
                GL.GL_FRAMEBUFFER,
                GL.GL_DEPTH_ATTACHMENT,
                GL.GL_TEXTURE_2D,
                self.db_tex.texture_name,
                0,  # mipmap level, normally 0
            )
            status = GL.glCheckFramebufferStatus(GL.GL_FRAMEBUFFER)
            if status != GL.GL_FRAMEBUFFER_COMPLETE:
                raise RuntimeError
            self.initialized = True
        GL.glBindFramebuffer(GL.GL_FRAMEBUFFER, self.fb_id)
        if clear:
            GL.glClear(GL.GL_COLOR_BUFFER_BIT | GL.GL_DEPTH_BUFFER_BIT)
        yield
        GL.glBindFramebuffer(GL.GL_FRAMEBUFFER, 0)

    @contextmanager
    def input_bind(self, fb_target=0, db_target=1):
        with self.fb_tex.bind(fb_target):
            with self.db_tex.bind(db_target):
                yield


class Texture3DIterator(traitlets.HasTraits):
    items = traitlets.Any()

    def __iter__(self, target=0):
        tex_target = TEX_TARGETS[target]
        for i, t in self.items:
            GL.glActiveTexture(tex_target)
            GL.glBindTexture(GL.GL_TEXTURE_3D, t.texture_name)
            yield i
        GL.glActiveTexture(tex_target)
        GL.glBindTexture(GL.GL_TEXTURE_3D, 0)


def compute_box_geometry(left_edge, right_edge):
    move = get_translate_matrix(*left_edge)
    width = right_edge - left_edge
    scale = get_scale_matrix(*width)

    transformed_box = bbox_vertices.dot(scale.T).dot(move.T).astype("float32")
    return transformed_box



import contextlib

import traitlets
from OpenGL import GL
from yt.data_objects.static_output import Dataset

from yt_idv.cameras.base_camera import BaseCamera
from yt_idv.cameras.trackball_camera import TrackballCamera
from yt_idv.opengl_support import Framebuffer
from yt_idv.scene_annotations.base_annotation import SceneAnnotation
from yt_idv.scene_annotations.box import BoxAnnotation
from yt_idv.scene_annotations.text import TextAnnotation
from yt_idv.scene_components.base_component import SceneComponent
from yt_idv.scene_components.blocks import BlockRendering
from yt_idv.scene_data.base_data import SceneData
from yt_idv.scene_data.block_collection import BlockCollection
from yt_idv.scene_data.box import BoxData
from yt_idv.scene_data.text_characters import TextCharacters


class SceneGraph(traitlets.HasTraits):
    components = traitlets.List(
        trait=traitlets.Instance(SceneComponent), default_value=[]
    )
    annotations = traitlets.List(
        trait=traitlets.Instance(SceneAnnotation), default_value=[]
    )
    data_objects = traitlets.List(trait=traitlets.Instance(SceneData), default_value=[])
    camera = traitlets.Instance(BaseCamera)
    ds = traitlets.Instance(Dataset)
    fb = traitlets.Instance(Framebuffer, allow_none=True)
    input_captured_mouse = traitlets.Bool(False)
    input_captured_keyboard = traitlets.Bool(False)

    def add_volume(self, data_source, field_name, no_ghost=False):
        """
        Add a BlockRendering component to volume render.

        This manages both creation of the block collection data and the
        rendering component associated with it.

        Parameters
        ----------
        data_source: yt data container such as a sphere, region, etc
        field_name: The field to volume render
        no_ghost: Should we save time by skipping ghost zone generation

        Returns
        -------

        component: BlockRendering

        """
        self.data_objects.append(BlockCollection(data_source=data_source))
        self.data_objects[-1].add_data(field_name, no_ghost=no_ghost)
        self.components.append(BlockRendering(data=self.data_objects[-1]))
        return self.components[-1]  # Only the rendering object

    def add_text(self, text="", origin=(0.0, 0.0), scale=1.0, **kwargs):
        """
        Add an annotation on top of the image in screen-space coordinates.

        This annotation object can also be updated in-place after the fact.

        Parameters
        ----------
        text: The text to draw
        origin: The location to place the origin of the text in screen-space
                coordinates (0..1)
        scale: The scale of the font

        Returns
        -------
        annotation: TextAnnotation

        Examples
        --------

        >>> sg.add_text("Hello there!", (0.2, 0.2))

        """
        if "data" not in kwargs:
            if not any(_.name == "text_overlay" for _ in self.data_objects):
                self.data_objects.append(TextCharacters())
                self.data_objects[-1].build_textures()
            kwargs["data"] = next(
                (_ for _ in self.data_objects if _.name == "text_overlay")
            )
        self.annotations.append(
            TextAnnotation(text=text, origin=origin, scale=scale, **kwargs)
        )
        return self.annotations[-1]

    def add_box(self, left_edge, right_edge):
        """
        Add a box annotation to the scene

        Parameters
        ----------
        left_edge: Array-like of the minimum of the box extent
        right_edge: Array-like of the maximum of the box extent
        """
        data = BoxData(left_edge=left_edge, right_edge=right_edge)
        self.data_objects.append(data)
        self.annotations.append(BoxAnnotation(data=data))

    def __iter__(self):
        """
        Iterate over all of the scene elements, first the components and then
        the annotations.

        Returns
        -------
        Iterator of the components and annotations sorted by priority.

        """
        elements = self.components + self.annotations
        for element in sorted(elements, key=lambda a: a.priority):
            yield element

    def render(self):
        """
        Render the scene into its local framebuffer.

        Returns
        -------
        Nothing, but the new image can be accessed.

        """
        origin_x, origin_y, width, height = GL.glGetIntegerv(GL.GL_VIEWPORT)
        with self.bind_buffer():
            GL.glClear(GL.GL_COLOR_BUFFER_BIT | GL.GL_DEPTH_BUFFER_BIT)
        for element in self:
            # If we need to clear the region, we need to use the scissor test
            db = element.display_bounds
            new_origin_x = origin_x + width * db[0]
            new_origin_y = origin_y + height * db[2]
            new_width = (db[1] - db[0]) * width
            new_height = (db[3] - db[2]) * height
            GL.glViewport(
                int(new_origin_x), int(new_origin_y), int(new_width), int(new_height)
            )
            if element.clear_region:
                GL.glEnable(GL.GL_SCISSOR_TEST)
                GL.glScissor(
                    int(new_origin_x),
                    int(new_origin_y),
                    int(new_width),
                    int(new_height),
                )
                GL.glClear(GL.GL_COLOR_BUFFER_BIT | GL.GL_DEPTH_BUFFER_BIT)
                GL.glDisable(GL.GL_SCISSOR_TEST)
            element.run_program(self)
        GL.glViewport(origin_x, origin_y, width, height)

    @contextlib.contextmanager
    def bind_buffer(self):
        """
        Provide a contextmanager, within which the local framebuffer is bound
        to the output target.

        Returns
        -------
        None

        """
        if self.fb is not None:
            with self.fb.bind(clear=False):
                yield
        else:
            yield

    @property
    def image(self):
        """
        This accesses the local framebuffer and returns the read pixels

        Returns
        -------
        RGBA array in floating point

        """
        if self.fb is not None:
            arr = self.fb.data[::-1, :, :]
            arr.swapaxes(0, 1)
            return arr
        _, _, width, height = GL.glGetIntegerv(GL.GL_VIEWPORT)
        arr = GL.glReadPixels(0, 0, width, height, GL.GL_RGBA, GL.GL_FLOAT)[::-1, :, :]
        arr.swapaxes(0, 1)
        return arr

    @property
    def depth(self):
        """
        This accesses the local depth buffer and returns the read pixels

        Returns
        -------
        Depth buffer in floating point format

        """
        if self.fb is not None:
            arr = self.fb.depth_data[::-1, :]
            arr.swapaxes(0, 1)
            return arr
        _, _, width, height = GL.glGetIntegerv(GL.GL_VIEWPORT)
        arr = GL.glReadPixels(0, 0, width, height, GL.GL_DEPTH_COMPONENT, GL.GL_FLOAT)[
            ::-1, :
        ]
        arr.swapaxes(0, 1)
        return arr

    @staticmethod
    def from_ds(ds, field, no_ghost=True):
        """
        Return a SceneGraph made from some best-guess values from a dataset and
        a field.

        This makes a couple assumptions, and given either a data source or a
        dataset, it will create a block rendering, jam a camera in there, and
        then return the scenegraph.

        Parameters
        ----------
        ds: A yt Dataset or Data Object to use as a source
        field: The field to render
        no_ghost: Should we save time by skipping ghost zones?

        Returns
        -------

        scene: SceneGraph

        Examples
        --------

        From a dataset:

        >>> sg = yt_idv.SceneGraph.from_ds(ds, "density")
        >>> rc = yt_idv.PygletRenderingContext(scene = sg)

        From a data object:

        >>> sp = ds.sphere("c", (0.1, 'unitary'))
        >>> sg = yt_idv.SceneGraph.from_ds(sp, "density")
        >>> rc = yt_idv.PygletRenderingContext(scene = sg)
        """
        if isinstance(ds, Dataset):
            data_source = ds.all_data()
        else:
            ds, data_source = ds.ds, ds
        center = ds.domain_center
        pos = center + 1.5 * ds.domain_width
        near_plane = 3.0 * ds.index.get_smallest_dx().min().d

        c = TrackballCamera(position=pos, focus=center, near_plane=near_plane)
        c.update_orientation(0, 0, 0, 0)

        scene = SceneGraph(camera=c)
        scene.add_volume(data_source, field, no_ghost=no_ghost)
        return scene



import os

def render_context(engine = "pyglet", **kwargs):
    """
    Return the appropriate rendering context.

    At present, this accepts either "pyglet" or "egl"

    Parameters
    ----------
    engine: str, either "pyglet" or "egl"

    Returns
    -------
    RenderingContext

    """
    if engine == "pyglet":
        from .pyglet_context import PygletRenderingContext
        return PygletRenderingContext(**kwargs)
    elif engine == "egl":
        os.environ["PYOPENGL_PLATFORM"] = "egl"
        from .egl_context import EGLRenderingContext
        return EGLRenderingContext(**kwargs)
    else:
        raise KeyError



#!/usr/bin/env python

"""The setup script."""

from setuptools import Extension, find_packages, setup  # NOQA
from Cython.Build import cythonize

with open("README.md") as readme_file:
    readme = readme_file.read()

with open("HISTORY.md") as history_file:
    history = history_file.read()

requirements = [
    "Click>=7.0",
    "yt>=4.0.dev0",
    "traitlets>=5.0.5",
    "pyOpenGL>=3.1.5",
    "traittypes>=0.2.1",
    "matplotlib>=3.0",
    "numpy>=1.18.0",
    "pyglet>=2.0.dev0",
    "pyyaml>=5.3.1",
    "imgui>=1.2.0",
]

setup_requirements = ["pytest-runner", "cython>=0.29"]

test_requirements = [
    "pytest>=3",
]

setup(
    author="Matthew Turk",
    author_email="matthewturk@gmail.com",
    python_requires=">=3.5",
    classifiers=[
        "Development Status :: 2 - Pre-Alpha",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: BSD License",
        "Natural Language :: English",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
    ],
    description="Interactive Volume Rendering for yt",
    entry_points={"console_scripts": ["yt_idv=yt_idv.cli:main",],},
    install_requires=requirements,
    license="BSD license",
    long_description=readme + "\n\n" + history,
    include_package_data=True,
    keywords="yt_idv",
    name="yt_idv",
    packages=find_packages(include=["yt_idv", "yt_idv.*"]),
    setup_requires=setup_requirements,
    test_suite="tests",
    tests_require=test_requirements,
    url="https://github.com/data-exp-lab/yt_idv",
    version="0.1.0",
    zip_safe=False,
    ext_modules=cythonize("yt_idv/*.pyx"),
)



import traitlets

from yt_idv.opengl_support import VertexArray, Texture


class SceneData(traitlets.HasTraits):
    """A class that defines a collection of GPU-managed data.

    This class contains the largest common set of features that can be used
    OpenGL rendering: a set of vertices and a set of vertex attributes.  Note
    that this is distinct from the shader, which can be swapped out and
    provided with these items.

    """

    name = None
    vertex_array = traitlets.Instance(VertexArray)
    textures = traitlets.List(trait=traitlets.Instance(Texture))

    min_val = traitlets.CFloat(0.0)
    max_val = traitlets.CFloat(1.0)
